# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Gives benchmark related info"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import jube2.util.util
import jube2.util.output
import jube2.conf
import jube2.jubeio
import os
import time
import textwrap
import operator


def print_benchmarks_info(path):
    """Print list of all benchmarks, found in given directory"""
    # Get list of all files and directories in given path
    if not os.path.isdir(path):
        raise OSError("Not a directory: \"{0}\"".format(path))
    dir_list = os.listdir(path)
    benchmark_info = list()
    # Search for possible benchmark dirs
    for dir_name in dir_list:
        dir_path = os.path.join(path, dir_name)
        configuration_file = \
            os.path.join(dir_path, jube2.conf.CONFIGURATION_FILENAME)
        if os.path.isdir(dir_path) and os.path.exists(configuration_file):
            try:
                id_number = int(dir_name)
                parser = jube2.jubeio.XMLParser(configuration_file)
                name_str, comment_str, tags = parser.benchmark_info_from_xml()
                tags_str = jube2.conf.DEFAULT_SEPARATOR.join(tags)

                # Read timestamps from timestamps file
                timestamps = \
                    jube2.util.util.read_timestamps(
                        os.path.join(dir_path, jube2.conf.TIMESTAMPS_INFO))
                if "start" in timestamps:
                    time_start = timestamps["start"]
                else:
                    time_start = time.strftime(
                        "%Y-%m-%d %H:%M:%S",
                        time.localtime(os.path.getctime(configuration_file)))
                if "change" in timestamps:
                    time_change = timestamps["change"]
                else:
                    time_change = time.strftime(
                        "%Y-%m-%d %H:%M:%S",
                        time.localtime(os.path.getmtime(dir_path)))

                benchmark_info.append([id_number, name_str, time_start,
                                       time_change, comment_str, tags_str])
            except ValueError:
                pass
    # sort using id
    benchmark_info = sorted(benchmark_info, key=operator.itemgetter(0))
    # convert id to string
    for info in benchmark_info:
        info[0] = str(info[0])
    # add header
    benchmark_info = [("id", "name", "started", "last change",
                       "comment", "tags")] + benchmark_info
    if len(benchmark_info) > 1:
        infostr = (jube2.util.output.text_boxed("Benchmarks found in \"{0}\":".
                                                format(path)) + "\n" +
                   jube2.util.output.text_table(benchmark_info,
                                                use_header_line=True))
        print(infostr)
    else:
        print("No Benchmarks found in \"{0}\"".format(path))


def print_benchmark_info(benchmark):
    """Print information concerning a single benchmark"""
    infostr = \
        jube2.util.output.text_boxed("{0} id:{1} tags:{2}\n\n{3}"
                                     .format(benchmark.name,
                                             benchmark.id,
                                             jube2.conf.DEFAULT_SEPARATOR.join(
                                                 benchmark.tags),
                                             benchmark.comment))
    print(infostr)
    continue_possible = False

    print("  Directory: {0}"
          .format(os.path.abspath(benchmark.bench_dir)))

    # Read timestamps from timestamps file
    timestamps = jube2.util.util.read_timestamps(
        os.path.join(benchmark.bench_dir, jube2.conf.TIMESTAMPS_INFO))

    if "start" in timestamps:
        time_start = timestamps["start"]
    else:
        # Starttime is workpackage.xml creation time
        time_start = time.strftime(
            "%Y-%m-%d %H:%M:%S",
            time.localtime(os.path.getctime(os.path.join(
                benchmark.bench_dir, jube2.conf.CONFIGURATION_FILENAME))))
    if "change" in timestamps:
        time_change = timestamps["change"]
    else:
        time_change = time.strftime(
            "%Y-%m-%d %H:%M:%S",
            time.localtime(os.path.getmtime(benchmark.bench_dir)))

    print("\n    Started: {0}".format(time_start))
    print("Last change: {0}".format(time_change))

    # Create step overview
    step_info = [("step name", "depends", "#work", "#error", "#done",
                  "last finished")]
    for step_name, workpackages in benchmark.workpackages.items():
        cnt_done = 0
        cnt_error = 0
        last_finish = time.localtime(0)
        depends = jube2.conf.DEFAULT_SEPARATOR.join(
            benchmark.steps[step_name].depend)
        for workpackage in workpackages:
            if workpackage.done:
                cnt_done += 1

                # Read timestamp from done_file if it is available otherwise
                # use mtime
                done_file = os.path.join(workpackage.workpackage_dir,
                                         jube2.conf.WORKPACKAGE_DONE_FILENAME)
                done_file_f = open(done_file, "r")
                done_str = done_file_f.read().strip()
                done_file_f.close()
                try:
                    done_time = time.strptime(done_str, "%Y-%m-%d %H:%M:%S")
                except ValueError:
                    done_time = time.localtime(os.path.getmtime(done_file))
                last_finish = max(last_finish, done_time)
            if workpackage.error:
                cnt_error += 1

        if last_finish > time.localtime(0):
            last_finish_str = time.strftime("%Y-%m-%d %H:%M:%S", last_finish)
        else:
            last_finish_str = ""
        continue_possible = continue_possible or \
            (len(workpackages) != cnt_done)

        # Create #workpackages string
        iterations = benchmark.steps[step_name].iterations
        if benchmark.steps[step_name].iterations > 1:
            cnt = "{0}*{1}".format(len(workpackages) // iterations, iterations)
        else:
            cnt = str(len(workpackages))

        step_info.append((step_name, depends, cnt, str(cnt_error),
                          str(cnt_done), last_finish_str))

    print(
        "\n" + jube2.util.output.text_table(step_info, use_header_line=True,
                                            indent=1))

    if continue_possible:
        print("\n--- Benchmark not finished! ---\n")
    else:
        print("\n--- Benchmark finished ---\n")

    print(jube2.util.output.text_line())


def print_step_info(benchmark, step_name, parametrization_only=False,
                    parametrization_only_csv=False):
    """Print information concerning a single step in a specific benchmark"""
    if step_name not in benchmark.workpackages:
        print("Step \"{0}\" not found in benchmark \"{1}\"."
              .format(step_name, benchmark.name))
        return

    if parametrization_only_csv:
        parametrization_only = True

    if not parametrization_only:
        print(jube2.util.output.text_boxed(
            "{0} Step: {1}".format(benchmark.name, step_name)))

    step = benchmark.steps[step_name]

    # Get all possible error filenames
    error_file_names = set()
    for operation in step.operations:
        if operation.stderr_filename is not None:
            error_file_names.add(operation.stderr_filename)
        else:
            error_file_names.add("stderr")

    wp_info = [("id", "started?", "error?", "done?", "work_dir")]
    error_dict = dict()
    parameter_list = list()
    useable_parameter = None
    for workpackage in benchmark.workpackages[step_name]:

        # Parameter substitution to use alt_work_dir
        parameter = \
            dict([[par.name, par.value] for par in
                  workpackage.parameterset.constant_parameter_dict.values()])

        # Save available parameter names
        if useable_parameter is None:
            useable_parameter = [name for name in parameter.keys()]
            useable_parameter.sort()

        id_str = str(workpackage.id)
        started_str = str(workpackage.started).lower()
        error_str = str(workpackage.error).lower()
        done_str = str(workpackage.done).lower()
        work_dir = workpackage.work_dir
        if step.alt_work_dir is not None:
            work_dir = jube2.util.util.substitution(step.alt_work_dir,
                                                    parameter)

        # collect parameterization
        parameter_list.append(dict())
        parameter_list[-1]["id"] = str(workpackage.id)
        for parameter in workpackage.parameterset:
            parameter_list[-1][parameter.name] = parameter.value

        # Read error-files
        for error_file_name in error_file_names:
            if os.path.exists(os.path.join(work_dir, error_file_name)):
                error_file = open(os.path.join(work_dir, error_file_name), "r")
                error_string = error_file.read().strip()
                if len(error_string) > 0:
                    error_dict[os.path.abspath(os.path.join(
                        work_dir, error_file_name))] = error_string
                error_file.close()

        # Store info data
        wp_info.append(
            (id_str, started_str, error_str, done_str,
             os.path.abspath(work_dir)))

    if not parametrization_only:
        print("Workpackages:")
        print(jube2.util.output.text_table(wp_info, use_header_line=True,
                                           indent=1, auto_linebreak=False))

    if (useable_parameter is not None) and (not parametrization_only):
        print("Available parameter:")
        wraps = textwrap.wrap(", ".join(useable_parameter), 80)
        for wrap in wraps:
            print(wrap)
        print("")

    if not parametrization_only:
        print("Parameterization:")
        for parameter_dict in parameter_list:
            print(" ID: {0}".format(parameter_dict["id"]))
            for name, value in parameter_dict.items():
                if name != "id":
                    print("  {0}: {1}".format(name, value))
            print()
    else:
        # Create parameterization table
        table_data = list()
        table_data.append(list())
        table_data[0].append("id")
        if len(parameter_list) > 0:
            for name in parameter_list[0]:
                if name != "id":
                    table_data[0].append(name)
            for parameter_dict in parameter_list:
                table_data.append(list())
                for name in table_data[0]:
                    table_data[-1].append(parameter_dict[name])
        print(jube2.util.output.text_table(
            table_data, use_header_line=True, indent=1, align_right=True,
            auto_linebreak=False, pretty=not parametrization_only_csv,
            separator=(parametrization_only_csv if (parametrization_only_csv)
                       else None)))

    if not parametrization_only:
        if len(error_dict) > 0:
            print("!!! Errors found !!!:")
        for error_file in error_dict:
            print(">>> {0}:".format(error_file))
            try:
                print("{0}\n".format(error_dict[error_file]))
            except UnicodeDecodeError:
                print("\n")


def print_benchmark_status(benchmark):
    """Print FINISHED or "RUNNING" dependign on the workpackage status"""
    all_done = True
    for step_name in benchmark.workpackages:
        for workpackage in benchmark.workpackages[step_name]:
            all_done = workpackage.done and all_done
    if all_done:
        print("FINISHED")
    else:
        print("RUNNING")



# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""The Workpackage class handles a step and its parameter space"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import xml.etree.ElementTree as ET
import jube2.util.util
import jube2.util.output
import jube2.conf
import jube2.log
import jube2.parameter
import os
import stat

LOGGER = jube2.log.get_logger(__name__)


class Workpackage(object):

    """A Workpackage contains all information to run a specific step with
    its given parameterset.
    """

    # class based counter for unique id creation
    id_counter = 0

    def __init__(self, benchmark, step, local_parameter_names, parameterset,
                 workpackage_id=None, iteration=0, cycle=0):
        # set id
        if workpackage_id is None:
            self._id = Workpackage.id_counter
            Workpackage.id_counter = Workpackage.id_counter + 1
        else:
            self._id = workpackage_id

        self._benchmark = benchmark
        self._step = step
        self._local_parameter_names = local_parameter_names
        self._parameterset = parameterset
        self._iteration = iteration
        self._parents = list()
        self._children = list()
        self._iteration_siblings = set()
        self._queued = False
        self._env = dict(os.environ)
        self._cycle = cycle
        self._workpackage_dir_caching_enabled = False
        self._workpackage_dir_cache = None

    def etree_repr(self):
        """Return etree object representation"""
        workpackage_etree = ET.Element("workpackage")
        workpackage_etree.attrib["id"] = str(self._id)
        step_etree = ET.SubElement(workpackage_etree, "step")
        step_etree.attrib["iteration"] = str(self._iteration)
        step_etree.attrib["cycle"] = str(self._cycle)
        step_etree.text = self._step.name
        if len(self._local_parameter_names) > 0:
            workpackage_etree.append(
                self.local_parameterset.etree_repr(use_current_selection=True))
        if len(self._parents) > 0:
            parents_etree = ET.SubElement(workpackage_etree, "parents")
            parents_etree.text = ",".join(
                [str(parent.id) for parent in self._parents])
        if len(self._iteration_siblings) > 0:
            sibling_etree = ET.SubElement(workpackage_etree,
                                          "iteration_siblings")
            sibling_etree.text = ",".join(
                [str(sibling.id) for sibling in self._iteration_siblings])
        environment_etree = ET.SubElement(workpackage_etree, "environment")
        for env_name, value in self._env.items():
            if (env_name not in ["PWD", "OLDPWD", "_"]) and \
                    (env_name not in os.environ or
                     os.environ[env_name] != value):
                env_etree = ET.SubElement(environment_etree, "env")
                env_etree.attrib["name"] = env_name
                # use string repr to avoid special characters
                env_etree.text = repr(value)
        for env_name in os.environ:
            if (env_name not in ["PWD", "OLDPWD", "_"]) and \
                    (env_name not in self._env):
                env_etree = ET.SubElement(environment_etree, "nonenv")
                env_etree.attrib["name"] = env_name
        return workpackage_etree

    def __repr__(self):
        return (("Workpackage(Id:{0:2d}; Step:{1}; ParentIDs:{2}; " +
                 "ChildIDs:{3} {4})").
                format(self._id, self._step.name,
                       [parent.id for parent in self._parents],
                       [child.id for child in self._children],
                       self.local_parameterset))

    def __eq__(self, other):
        if isinstance(other, Workpackage):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        return object.__hash__(self)

    @property
    def parameter_dict(self):
        """get all available parameter inside a dict"""
        # Collect parameter for substitution
        parameter = dict([[par.name, par.value] for par in
                          self._parameterset.constant_parameter_dict.values()])
        return parameter

    @property
    def env(self):
        """Return workpackage environment"""
        return self._env

    @property
    def cycle(self):
        """Return current loop cycle"""
        return self._cycle

    def allow_workpackage_dir_caching(self):
        """Enable workpackage dir cache"""
        self._workpackage_dir_caching_enabled = True
        self._workpackage_dir_cache = None

    @property
    def active(self):
        """Check active state"""
        active = self._step.active
        # Collect parameter for substitution
        parameter = self.parameter_dict
        # Parameter substitution
        active = jube2.util.util.substitution(active, parameter)
        # Evaluate active state
        return jube2.util.util.eval_bool(active)

    @property
    def done(self):
        """Workpackage done?"""
        done_file = os.path.join(self.workpackage_dir,
                                 jube2.conf.WORKPACKAGE_DONE_FILENAME)
        exist = os.path.exists(done_file)
        if jube2.conf.DEBUG_MODE:
            exist = exist or os.path.exists(done_file + "_DEBUG")
        return exist

    @done.setter
    def done(self, set_done):
        """Set/reset Workpackage done"""
        done_file = os.path.join(self.workpackage_dir,
                                 jube2.conf.WORKPACKAGE_DONE_FILENAME)
        if jube2.conf.DEBUG_MODE:
            done_file = done_file + "_DEBUG"
        if set_done:
            fout = open(done_file, "w")
            fout.write(jube2.util.util.now_str())
            fout.close()
            self._remove_operation_info_files()
        else:
            if os.path.exists(done_file):
                os.remove(done_file)

    @property
    def error(self):
        """Workpackage error?"""
        error_file = os.path.join(self.workpackage_dir,
                                  jube2.conf.WORKPACKAGE_ERROR_FILENAME)
        return os.path.exists(error_file)

    def set_error(self, set_error, msg=""):
        """Set/reset Workpackage error"""
        error_file = os.path.join(self.workpackage_dir,
                                  jube2.conf.WORKPACKAGE_ERROR_FILENAME)
        if set_error:
            fout = open(error_file, "w")
            fout.write(msg)
            fout.close()
        else:
            if os.path.exists(error_file):
                os.remove(error_file)

    @property
    def queued(self):
        """Workpackage queued?"""
        return self._queued

    @queued.setter
    def queued(self, set_queued):
        """Set queued state"""
        self._queued = set_queued

    @property
    def started(self):
        """Workpackage started?"""
        return os.path.exists(self.workpackage_dir)

    def operation_done_but_pending(self, operation_number):
        """Check if an operation was executed, but the result is still
        pending (because it is a async do)"""
        result = self.operation_done(operation_number)

        operation = self._step.operations[operation_number]

        if result and (operation.async_filename is not None):
            parameter_dict = self.parameter_dict
            if operation.active(parameter_dict):
                work_dir = self.work_dir
                alt_work_dir = self.alt_work_dir(parameter_dict)
                if alt_work_dir is not None:
                    work_dir = alt_work_dir
                async_filename = jube2.util.util.substitution(
                    operation.async_filename, parameter_dict)
                async_filename = \
                    os.path.expandvars(os.path.expanduser(async_filename))
                result = not os.path.exists(os.path.join(work_dir,
                                                         async_filename))
            else:
                result = False
        else:
            result = False
        return result

    def operation_done(self, operation_number, set_done=None):
        """Mark/checks operation status"""
        done_file = os.path.join(self.workpackage_dir,
                                 "wp_{0}_{1:02d}".format(
                                     jube2.conf.WORKPACKAGE_DONE_FILENAME,
                                     operation_number))
        if set_done is None:
            exist = os.path.exists(done_file)
            if jube2.conf.DEBUG_MODE:
                exist = exist or os.path.exists(done_file + "_DEBUG")
            return exist
        else:
            if jube2.conf.DEBUG_MODE:
                done_file = done_file + "_DEBUG"
            elif ((set_done and not os.path.exists(done_file)) or
                  (not set_done and os.path.exists(done_file))):
                jube2.util.util.update_timestamps(
                    os.path.join(self._benchmark.bench_dir,
                                 jube2.conf.TIMESTAMPS_INFO),
                    "change")
            if set_done:
                fout = open(done_file, "w")
                fout.close()
            else:
                if os.path.exists(done_file):
                    os.remove(done_file)
            return set_done

    def _remove_operation_info_files(self):
        """Remove all operation info files"""
        for operation_number in range(len(self._step.operations)):
            self.operation_done(operation_number, False)

    def add_parent(self, workpackage):
        """Add a parent Workpackage"""
        self._parents.append(workpackage)

    @property
    def parameterset(self):
        """Return parameterset"""
        return self._parameterset

    def add_children(self, workpackage):
        """Add a children workpackage"""
        self._children.append(workpackage)

    @property
    def local_parameterset(self):
        """Return local parameterset"""
        parameterset = jube2.parameter.Parameterset()
        for name in self._local_parameter_names:
            parameterset.add_parameter(self._parameterset[name])
        return parameterset

    @property
    def parent_history(self):
        """Create a list of all parents in the history of this workpackage"""
        history = list()
        for parent in self._parents:
            history += parent.parent_history
        history += self._parents
        return history

    @property
    def children_future(self):
        """Create a list of all children in the future of this workpackage"""
        future = list()
        future += self._children
        for child in self._children:
            future += child.children_future
        return future

    @property
    def id(self):
        """Return workpackage id"""
        return self._id

    @property
    def parents(self):
        """Return list of parent workpackages"""
        return self._parents

    @property
    def iteration_siblings(self):
        """Return set of iteration siblings"""
        return self._iteration_siblings

    @property
    def iteration(self):
        """Return workpackage iteration number"""
        return self._iteration

    @property
    def children(self):
        """Return list of child workpackages"""
        return self._children

    @property
    def step(self):
        """Return Step data"""
        return self._step

    def get_jube_cycle_parameterset(self):
        """Return parameterset which contains cycle related
        information"""
        parameterset = jube2.parameter.Parameterset()

        # worpackage cycle
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_cycle",
                             str(self._cycle), parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        return parameterset

    def create_relpath(self, value):
        """Create relative path representation"""
        return os.path.relpath(
            os.path.join(self._benchmark.file_path_ref, value),
            self._benchmark.file_path_ref)

    def create_abspath(self, value):
        """Create absolute path representation"""
        return os.path.abspath(
            os.path.join(self._benchmark.file_path_ref, value))

    def get_jube_parameterset(self):
        """Return parameterset which contains workpackage related
        information"""
        parameterset = jube2.parameter.Parameterset()
        # workpackage id
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_id", str(self._id),
                             parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        # workpackage id with padding
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_padid",
                             jube2.util.util.id_dir("", self._id),
                             parameter_type="string",
                             update_mode=jube2.parameter.JUBE_MODE))

        # workpackage iteration
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_iteration",
                             str(self._iteration), parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        parameterset.add_parameterset(self.get_jube_cycle_parameterset())

        # pathes
        if self._step.alt_work_dir is None:
            path = self.work_dir
        else:
            path = self._step.alt_work_dir

        # workpackage relative folder path
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_relpath", path,
                             update_mode=jube2.parameter.JUBE_MODE,
                             eval_helper=self.create_relpath))

        # workpackage absolute folder path
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_abspath", path,
                             update_mode=jube2.parameter.JUBE_MODE,
                             eval_helper=self.create_abspath))

        # parent workpackage id
        for parent in self._parents:
            parameterset.add_parameter(
                jube2.parameter.Parameter.
                create_parameter(("jube_wp_parent_{0}_id")
                                 .format(parent.step.name),
                                 str(parent.id), parameter_type="int",
                                 update_mode=jube2.parameter.JUBE_MODE))

        # environment export string
        env_str = ""
        parameter_names = [parameter.name for parameter in
                           self._parameterset.export_parameter_dict.values()]
        parameter_names.sort(key=str.lower)
        for name in parameter_names:
            env_str += "export {0}=${1}\n".format(name, name)
        env_par = jube2.parameter.Parameter.create_parameter(
            "jube_wp_envstr", env_str, no_templates=True,
            update_mode=jube2.parameter.JUBE_MODE,
            eval_helper=jube2.parameter.StaticParameter.fix_export_string)
        parameterset.add_parameter(env_par)

        # environment export list
        parameterset.add_parameter(
            jube2.parameter.Parameter.create_parameter(
                "jube_wp_envlist",
                ",".join([name for name in parameter_names]),
                no_templates=True, update_mode=jube2.parameter.JUBE_MODE))

        return parameterset

    def create_workpackage_dir(self):
        """Create work directory"""
        if not os.path.exists(self.workpackage_dir):
            if "$" in self.workpackage_dir:
                raise RuntimeError(("'{0}' could not be evaluated and used " +
                                    "as a workpackage directory name. " +
                                    "Please check the suffix setting.")
                                   .format(self.workpackage_dir))
            os.mkdir(self.workpackage_dir)
            os.mkdir(self.work_dir)

        # Create symbolic link to parent workpackage folder
        for parent in self._parents:
            link_path = os.path.join(self.work_dir, parent.step.name)
            parent_path = os.path.relpath(parent.work_dir, self.work_dir)
            if not os.path.exists(link_path):
                os.symlink(parent_path, link_path)

    def create_shared_folder_link(self, parameter_dict=None):
        """Create shared folder connection"""
        # Create symbolic link to shared folder
        if self._step.shared_link_name is not None:
            shared_folder = self._step.shared_folder_path(
                self._benchmark.bench_dir, parameter_dict)
            # Create shared folder (if it not already exists)
            if not os.path.exists(shared_folder):
                os.mkdir(shared_folder)

            # Create shared folder link
            if parameter_dict is not None:
                shared_name = \
                    jube2.util.util.substitution(self._step.shared_link_name,
                                                 parameter_dict)
            else:
                shared_name = self._step.shared_link_name
            link_path = os.path.join(self.work_dir, shared_name)
            target_path = \
                os.path.relpath(shared_folder, self.work_dir)
            if not os.path.exists(link_path):
                os.symlink(target_path, link_path)

    @property
    def workpackage_dir(self):
        """Return workpackage directory"""
        if not self._workpackage_dir_caching_enabled or \
                self._workpackage_dir_cache is None:
            suffix = self.step.suffix
            if suffix != "":
                # Collect parameter for substitution
                parameter = \
                    dict([[par.name, par.value] for par in
                          self._parameterset.constant_parameter_dict.values()])
                # Parameter substitution
                suffix = jube2.util.util.substitution(suffix, parameter)
                suffix = "_" + os.path.expandvars(os.path.expanduser(suffix))
            path = "{path}_{step_name}{suffix}".format(
                path=jube2.util.util.id_dir(
                    self._benchmark.bench_dir, self._id),
                step_name=self._step.name,
                suffix=suffix)
        if self._workpackage_dir_caching_enabled:
            if self._workpackage_dir_cache is None:
                self._workpackage_dir_cache = path
            return self._workpackage_dir_cache
        else:
            return path

    @property
    def work_dir(self):
        """Return working directory (user space)"""
        return os.path.join(self.workpackage_dir, "work")

    def alt_work_dir(self, parameter_dict=None):
        """Return location of alternative working_dir"""
        if self._step.alt_work_dir is not None:
            if parameter_dict is None:
                parameter_dict = self.parameter_dict
            alt_work_dir = self._step.alt_work_dir
            alt_work_dir = jube2.util.util.substitution(alt_work_dir,
                                                        parameter_dict)
            alt_work_dir = os.path.expandvars(os.path.expanduser(alt_work_dir))
            alt_work_dir = os.path.join(self._benchmark.file_path_ref,
                                        alt_work_dir)
            return alt_work_dir
        else:
            return None

    def _run_operations(self, parameter, work_dir):
        """Run all available operations"""
        continue_op = True
        continue_cycle = True
        for operation_number, operation in enumerate(self._step.operations):
            # Check if the operation is activated
            active = operation.active(parameter)
            if not active:
                self.operation_done(operation_number, True)
            # Do nothing, if the next operation is already finished.
            # Otherwise a removed async_file will result in a new
            # pending operation, if there are two async-operations in
            # a row
            elif not self.operation_done(operation_number + 1):
                # shared operation
                if operation.shared:
                    # wait for all other workpackages and check if shared
                    # operation already finished
                    shared_done = False
                    for workpackage in \
                            self._benchmark.workpackages[self._step.name]:
                        # All workpackages must reach the same position in
                        # the program
                        if operation_number > 0:
                            continue_op = continue_op and \
                                ((workpackage.operation_done(
                                    operation_number - 1) and
                                  (not workpackage.operation_done_but_pending(
                                      operation_number - 1))
                                  ) or
                                 workpackage.done) and \
                                workpackage.cycle == self._cycle
                        # Check if another workpackage already finalized
                        # the operation, only if the operation was active
                        # for this particular workpackage
                        shared_done = shared_done or \
                            ((workpackage.operation_done(
                                operation_number + 1) or workpackage.done
                              ) and
                             operation.active(workpackage.parameter_dict))

                    # All older workpackages in tree must be done
                    for step_name in self._step.get_depend_history(
                            self._benchmark):
                        for workpackage in self._benchmark.workpackages[
                                step_name]:
                            continue_op = continue_op and workpackage.done

                    if continue_op and not shared_done:
                        # remove workpackage specific parameter
                        shared_parameter = dict(parameter)
                        for jube_parameter in self.get_jube_parameterset()\
                                .all_parameter_names:
                            if jube_parameter in shared_parameter:
                                del shared_parameter[jube_parameter]

                        # work_dir = shared_dir
                        shared_dir = \
                            self._step.shared_folder_path(
                                self._benchmark.bench_dir, shared_parameter)

                        LOGGER.debug("====== {0} - shared ======"
                                     .format(self._step.name))

                        continue_op, continue_cycle = operation.execute(
                            parameter_dict=shared_parameter,
                            work_dir=shared_dir,
                            environment=self._env,
                            only_check_pending=self.operation_done(
                                operation_number))

                        # update all workpackages
                        for workpackage in self._benchmark.workpackages[
                                self._step.name]:
                            # if the operation wasn't active in the shared
                            # operation it must not be triggered to
                            # restart
                            if operation.active(
                                    workpackage.parameter_dict):
                                if not workpackage.started:
                                    workpackage.create_workpackage_dir()
                                workpackage.operation_done(
                                    operation_number, True)
                                if continue_op and not continue_cycle:
                                    workpackage.done = True
                                # requeue other workpackages
                                if not workpackage.queued and continue_op:
                                    self._benchmark.work_stat.put(
                                        workpackage)
                        LOGGER.debug("======================={0}"
                                     .format(len(self._step.name) * "="))
                else:
                    continue_op, continue_cycle = operation.execute(
                        parameter_dict=parameter, work_dir=work_dir,
                        environment=self._env,
                        only_check_pending=self.operation_done(
                            operation_number))
                    self.operation_done(operation_number, True)
            if not continue_op or not continue_cycle:
                break
        return continue_op, continue_cycle

    def run(self):
        """Run step and use current parameter space"""

        # Workpackage already done or error?
        if self.done or self.error:
            return

        continue_op = True
        continue_cycle = True
        while (continue_cycle and continue_op):

            stepstr = ("{0} ( iter:{2} | id:{1} | parents:{3} | cycle:{4} )"
                       .format(self._step.name, self._id, self._iteration,
                               ",".join([parent.step.name + "(" +
                                         str(parent.id) + ")"
                                         for parent in self._parents]),
                               self._cycle))
            stepstr = "----- {0} -----".format(stepstr)
            LOGGER.debug(stepstr)

            # --- Check if this is the first run ---
            started_before = self.started

            # --- Create directory structure ---
            if not started_before:
                self.create_workpackage_dir()

            # --- Load environment of parent steps ---
            if not started_before:
                for parent in self._parents:
                    if parent.step.export:
                        self._env.update(parent.env)

            # --- Update JUBE parameter for new cycle ---
            if self._cycle > 0:
                self.parameterset.update_parameterset(
                    self.get_jube_cycle_parameterset())

            # --- Update cycle parameter ---
            update_parameter = \
                self.parameterset.get_updatable_parameter(
                    mode=jube2.parameter.CYCLE_MODE, keep_index=True)
            if len(update_parameter) > 0:
                fixed_parameterset = self.parameterset.copy()
                for parameter in update_parameter:
                    fixed_parameterset.delete_parameter(parameter)
                change = True
                while change:
                    change = False
                    update_parameter.parameter_substitution(
                        [fixed_parameterset])
                    if update_parameter.has_templates:
                        update_parameter = list(
                            update_parameter.expand_templates())[0]
                        change = True
                update_parameter.parameter_substitution(
                    [fixed_parameterset], final_sub=True)
                self.parameterset.update_parameterset(update_parameter)
                debugstr = "  updated parameter:\n"
                debugstr += jube2.util.output.text_table(
                    [("parameter", "value")] + sorted(
                        [(par.name, par.value) for par in update_parameter]),
                    use_header_line=True, indent=9, align_right=False)
                LOGGER.debug(debugstr)

            # --- Collect parameter for substitution ---
            parameter = self.parameter_dict

            if not started_before:
                # --- Collect export parameter ---
                self._env.update(
                    dict([[par.name, par.value] for par in
                          self._parameterset.export_parameter_dict.values()]))

            # --- Create shared folder connection ---
            if self._cycle == 0:
                self.create_shared_folder_link(parameter)

            # --- Create alternativ working dir ---
            alt_work_dir = self.alt_work_dir(parameter)
            if alt_work_dir is not None:
                LOGGER.debug("  switch to alternativ work dir: \"{0}\""
                             .format(alt_work_dir))
                if not jube2.conf.DEBUG_MODE and \
                        not os.path.exists(alt_work_dir):
                    os.makedirs(alt_work_dir)
                    # Get group_id if available (given by JUBE_GROUP_NAME)
                    group_id = jube2.util.util.check_and_get_group_id()
                    if group_id is not None:
                        os.chown(alt_work_dir, os.getuid(), group_id)
                        os.chmod(alt_work_dir,
                                 os.stat(alt_work_dir).st_mode | stat.S_ISGID)

            # Print debug info
            if self._cycle == 0:
                debugstr = "  available parameter:\n"
                debugstr += jube2.util.output.text_table(
                    [("parameter", "value")] + sorted(
                        [(name, par) for name, par in parameter.items()]),
                    use_header_line=True, indent=9,
                    align_right=False)
                LOGGER.debug(debugstr)

            # --- Copy files to working dir or create links ---
            if not started_before:
                # Filter for filesets in uses
                fileset_names = \
                    self._step.get_used_sets(self._benchmark.filesets,
                                             parameter)
                for name in fileset_names:
                    self._benchmark.filesets[name].create(
                        work_dir=self.work_dir,
                        parameter_dict=parameter,
                        alt_work_dir=alt_work_dir,
                        environment=self._env,
                        file_path_ref=self._benchmark.file_path_ref)

            work_dir = self.work_dir
            if alt_work_dir is not None:
                work_dir = alt_work_dir

            # --- File substitution ---
            if not started_before:
                # Filter for substitutionsets in uses
                substituteset_names = \
                    self._step.get_used_sets(self._benchmark.substitutesets,
                                             parameter)
                for name in substituteset_names:
                    self._benchmark.substitutesets[name].substitute(
                        parameter_dict=parameter, work_dir=work_dir)

            try:
                # Run all operations
                # continue_op = false means -> async operation or wait for
                #     others in shared operation
                # continue_cycle = false -> loop cycle was interrupted
                continue_op, continue_cycle = \
                    self._run_operations(parameter, work_dir)

                # --- Check cycle limit ---
                if self._cycle + 1 >= self._step.cycles:
                    continue_cycle = False

                if continue_op and continue_cycle:
                    # --- Prepare additional cycle if needed ---
                    self._cycle += 1
                    self._remove_operation_info_files()
                elif continue_op:
                    # --- Write information file to mark end of work ---
                    self.done = True
            except RuntimeError as re:
                self.set_error(True, str(re))
                continue_cycle = False
                if jube2.conf.EXIT_ON_ERROR:
                    raise(RuntimeError(str(re)))
                else:
                    LOGGER.debug(
                        "{0}\n{1}\n{2}".format(40 * "-", str(re), 40 * "-"))

    @staticmethod
    def reduce_workpackage_id_counter():
        Workpackage.id_counter = Workpackage.id_counter - 1



# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Step contains the commands for steps"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import subprocess
import os
import re
import time
import xml.etree.ElementTree as ET
import jube2.util.util
import jube2.conf
import jube2.log

LOGGER = jube2.log.get_logger(__name__)


class Step(object):

    """A Step represent one execution step. It contains a list of
    Do-operations and multiple parametersets, substitutionsets and filesets.
    A Step is a template for Workpackages.
    """

    def __init__(self, name, depend, iterations=1, alt_work_dir=None,
                 shared_name=None, export=False, max_wps="0",
                 active="true", suffix="", cycles=1):
        self._name = name
        self._use = list()
        self._operations = list()
        self._iterations = iterations
        self._depend = depend
        self._alt_work_dir = alt_work_dir
        self._shared_name = shared_name
        self._export = export
        self._max_wps = max_wps
        self._active = active
        self._suffix = suffix
        self._cycles = cycles

    def etree_repr(self):
        """Return etree object representation"""
        step_etree = ET.Element("step")
        step_etree.attrib["name"] = self._name
        if len(self._depend) > 0:
            step_etree.attrib["depend"] = \
                jube2.conf.DEFAULT_SEPARATOR.join(self._depend)
        if self._alt_work_dir is not None:
            step_etree.attrib["work_dir"] = self._alt_work_dir
        if self._shared_name is not None:
            step_etree.attrib["shared"] = self._shared_name
        if self._active != "true":
            step_etree.attrib["active"] = self._active
        if self._suffix != "":
            step_etree.attrib["suffix"] = self._suffix
        if self._export:
            step_etree.attrib["export"] = "true"
        if self._max_wps != "0":
            step_etree.attrib["max_async"] = self._max_wps
        if self._iterations > 1:
            step_etree.attrib["iterations"] = str(self._iterations)
        if self._cycles > 1:
            step_etree.attrib["cycles"] = str(self._cycles)
        for use in self._use:
            use_etree = ET.SubElement(step_etree, "use")
            use_etree.text = jube2.conf.DEFAULT_SEPARATOR.join(use)
        for operation in self._operations:
            step_etree.append(operation.etree_repr())
        return step_etree

    def __repr__(self):
        return "{0}".format(vars(self))

    def add_operation(self, operation):
        """Add operation"""
        self._operations.append(operation)

    def add_uses(self, use_names):
        """Add use"""
        for use_name in use_names:
            if any([use_name in use_list for use_list in self._use]):
                raise ValueError(("Element \"{0}\" can only be used once")
                                 .format(use_name))
        self._use.append(use_names)

    @property
    def name(self):
        """Return step name"""
        return self._name

    @property
    def active(self):
        """Return active state"""
        return self._active

    @property
    def export(self):
        """Return export behaviour"""
        return self._export

    @property
    def iterations(self):
        """Return iterations"""
        return self._iterations

    @property
    def cycles(self):
        """Return number of cycles"""
        return self._cycles

    @property
    def shared_link_name(self):
        """Return shared link name"""
        return self._shared_name

    @property
    def max_wps(self):
        """Return maximum number of simultaneous workpackages"""
        return self._max_wps

    def get_used_sets(self, available_sets, parameter_dict=None):
        """Get list of all used sets, which can be found in available_sets"""
        set_names = list()
        if parameter_dict is None:
            parameter_dict = dict()
        for use in self._use:
            for name in use:
                name = jube2.util.util.substitution(name, parameter_dict)
                if (name in available_sets) and (name not in set_names):
                    set_names.append(name)
        return set_names

    def shared_folder_path(self, benchdir, parameter_dict=None):
        """Return shared folder name"""
        if self._shared_name is not None:
            if parameter_dict is not None:
                shared_name = jube2.util.util.substitution(self._shared_name,
                                                           parameter_dict)
            else:
                shared_name = self._shared_name
            return os.path.join(benchdir,
                                "{0}_{1}".format(self._name, shared_name))
        else:
            return ""

    def get_jube_parameterset(self):
        """Return parameterset which contains step related
        information"""
        parameterset = jube2.parameter.Parameterset()

        # step name
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_step_name", self._name,
                             update_mode=jube2.parameter.JUBE_MODE))

        # iterations
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_step_iterations", str(self._iterations),
                             parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        # cycles
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_step_cycles", str(self._cycles),
                             parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        # default worpackage cycle, will be overwritten by specific worpackage
        # cycle
        parameterset.add_parameter(
            jube2.parameter.Parameter.
            create_parameter("jube_wp_cycle", "0", parameter_type="int",
                             update_mode=jube2.parameter.JUBE_MODE))

        return parameterset

    def create_workpackages(self, benchmark, global_parameterset,
                            local_parameterset=None, used_sets=None,
                            iteration_base=0, parents=None,
                            incompatible_parameters=None):
        """Create workpackages for current step using given
        benchmark context"""
        if used_sets is None:
            used_sets = set()

        update_parameters = jube2.parameter.Parameterset()
        if local_parameterset is None:
            local_parameterset = jube2.parameter.Parameterset()
            global_parameterset.add_parameterset(
                benchmark.get_jube_parameterset())
            global_parameterset.add_parameterset(self.get_jube_parameterset())
            update_parameters.add_parameterset(
                global_parameterset.get_updatable_parameter(
                    jube2.parameter.STEP_MODE))
            for parameter in update_parameters:
                incompatible_parameters.discard(parameter.name)

        if parents is None:
            parents = list()

        new_workpackages = list()

        # Create parameter dictionary for substitution
        parameter_dict = \
            dict([[par.name, par.value] for par in
                  global_parameterset.constant_parameter_dict.values()])

        # Filter for parametersets in uses
        parameterset_names = \
            set(self.get_used_sets(benchmark.parametersets, parameter_dict))
        new_sets_found = len(parameterset_names.difference(used_sets)) > 0

        if new_sets_found:
            parameterset_names = parameterset_names.difference(used_sets)
            used_sets = used_sets.union(parameterset_names)

            for parameterset_name in parameterset_names:
                # The parametersets in a single step must be compatible
                if not local_parameterset.is_compatible(
                        benchmark.parametersets[parameterset_name]):
                    incompatible_names = \
                        local_parameterset.get_incompatible_parameter(
                            benchmark.parametersets[parameterset_name])
                    raise ValueError(("Cannot use parameterset '{0}' in " +
                                      "step '{1}'.\nParameter '{2}' is/are " +
                                      "already defined by a different " +
                                      "parameterset.")
                                     .format(parameterset_name, self.name,
                                             ",".join(incompatible_names)))
                local_parameterset.add_parameterset(
                    benchmark.parametersets[parameterset_name])

            # Combine local and history parameterset
            if local_parameterset.is_compatible(
                    global_parameterset, update_mode=jube2.parameter.USE_MODE):
                update_parameters.add_parameterset(
                    local_parameterset.get_updatable_parameter(
                        jube2.parameter.USE_MODE))
                for parameter in update_parameters:
                    incompatible_parameters.discard(parameter.name)
                global_parameterset = \
                    local_parameterset.copy().add_parameterset(
                        global_parameterset)
            else:
                incompatible_names = \
                    local_parameterset.get_incompatible_parameter(
                        global_parameterset,
                        update_mode=jube2.parameter.USE_MODE)
                LOGGER.debug("Incompatible parameterset combination found " +
                             "between current and parent steps. \nParameter " +
                             "'{0}' is/are already defined different.".format(
                                 ",".join(incompatible_names)))
                return new_workpackages

        # update parameters
        global_parameterset.update_parameterset(update_parameters)

        # Expand templates
        parametersets = [global_parameterset]
        change = True
        while change:
            change = False
            new_parametersets = list()
            for parameterset in parametersets:
                parameterset.parameter_substitution()
                # Maybe new templates were created
                if parameterset.has_templates:
                    LOGGER.debug("Expand parameter templates:\n{0}".format(
                        "\n".join("    \"{0}\": {1}".format(i, j.value)
                                  for i, j in parameterset.
                                  template_parameter_dict.items())))
                    new_parametersets += \
                        [new_parameterset for new_parameterset in
                         parameterset.expand_templates()]
                    change = True
                else:
                    new_parametersets += [parameterset]
            parametersets = new_parametersets

        # Create workpackages
        for parameterset in parametersets:
            workpackage_parameterset = local_parameterset.copy()
            workpackage_parameterset.update_parameterset(parameterset)
            if new_sets_found:
                new_workpackages += \
                    self.create_workpackages(benchmark, parameterset,
                                             workpackage_parameterset,
                                             used_sets, iteration_base,
                                             parents,
                                             incompatible_parameters.copy())
            else:
                # Check if all incompatible_parameters were updated
                if len(incompatible_parameters) > 0:
                    return new_workpackages
                # Create new workpackage
                created_workpackages = list()
                for iteration in range(self.iterations):
                    workpackage = jube2.workpackage.Workpackage(
                        benchmark=benchmark,
                        step=self,
                        parameterset=parameterset.copy(),
                        local_parameter_names=[
                            par.name for par in workpackage_parameterset],
                        iteration=iteration_base * self.iterations + iteration,
                        cycle=0)

                    # --- Link parent workpackages ---
                    for parent in parents:
                        workpackage.add_parent(parent)

                    # --- Add workpackage JUBE parameterset ---
                    workpackage.parameterset.add_parameterset(
                        workpackage.get_jube_parameterset())

                    # --- Final parameter substitution ---
                    workpackage.parameterset.parameter_substitution(
                        final_sub=True)

                    # --- Check parameter type ---
                    for parameter in workpackage.parameterset:
                        if not parameter.is_template:
                            jube2.util.util.convert_type(
                                parameter.parameter_type, parameter.value)

                    # --- Enable workpackage dir cache ---
                    workpackage.allow_workpackage_dir_caching()

                    if workpackage.active:
                        created_workpackages.append(workpackage)
                    else:
                        jube2.workpackage.Workpackage.\
                            reduce_workpackage_id_counter()

                for workpackage in created_workpackages:
                    workpackage.iteration_siblings.update(
                        set(created_workpackages))

                new_workpackages += created_workpackages

        return new_workpackages

    @property
    def alt_work_dir(self):
        """Return alternativ work directory"""
        return self._alt_work_dir

    @property
    def use(self):
        """Return parameters and substitutions"""
        return self._use

    @property
    def suffix(self):
        """Return directory suffix"""
        return self._suffix

    @property
    def operations(self):
        """Return operations"""
        return self._operations

    @property
    def depend(self):
        """Return dependencies"""
        return self._depend

    def get_depend_history(self, benchmark):
        """Creates a set of all dependent steps in history for given
        benchmark"""
        depend_history = set()
        for step_name in self._depend:
            if step_name not in depend_history:
                depend_history.add(step_name)
                depend_history.update(
                    benchmark.steps[step_name].get_depend_history(benchmark))
        return depend_history


class Operation(object):

    """The Operation-class represents a single instruction, which will be
    executed in a shell environment.
    """

    def __init__(self, do, async_filename=None, stdout_filename=None,
                 stderr_filename=None, active="true", shared=False,
                 work_dir=None, break_filename=None):
        self._do = do
        self._async_filename = async_filename
        self._break_filename = break_filename
        self._stdout_filename = stdout_filename
        self._stderr_filename = stderr_filename
        self._active = active
        self._shared = shared
        self._work_dir = work_dir

    @property
    def stdout_filename(self):
        """Get stdout filename"""
        return self._stdout_filename

    @property
    def stderr_filename(self):
        """Get stderr filename"""
        return self._stderr_filename

    @property
    def async_filename(self):
        """Get async filename"""
        return self._async_filename

    @property
    def shared(self):
        """Shared operation?"""
        return self._shared

    def active(self, parameter_dict):
        """Return active status of the current operation depending on the
        given parameter_dict"""
        active_str = jube2.util.util.substitution(self._active, parameter_dict)
        return jube2.util.util.eval_bool(active_str)

    def execute(self, parameter_dict, work_dir, only_check_pending=False,
                environment=None):
        """Execute the operation. work_dir must be set to the given context
        path. The parameter_dict used for inline substitution.
        If only_check_pending is set to True, the operation will not be
        executed, only the async_file will be checked.
        Return operation status:
        True => operation finished
        False => operation pending
        """
        if not self.active(parameter_dict):
            return True

        if environment is not None:
            env = environment
        else:
            env = os.environ

        if not only_check_pending:
            # Inline substitution
            do = jube2.util.util.substitution(self._do, parameter_dict)

            # Remove leading and trailing ; because otherwise ;; will cause
            # trouble when adding ; env
            do = do.strip(";")

            if (not jube2.conf.DEBUG_MODE) and (do.strip() != ""):
                # Change stdout
                if self._stdout_filename is not None:
                    stdout_filename = jube2.util.util.substitution(
                        self._stdout_filename, parameter_dict)
                    stdout_filename = \
                        os.path.expandvars(os.path.expanduser(stdout_filename))
                else:
                    stdout_filename = "stdout"
                stdout_path = os.path.join(work_dir, stdout_filename)
                stdout = open(stdout_path, "a")

                # Change stderr
                if self._stderr_filename is not None:
                    stderr_filename = jube2.util.util.substitution(
                        self._stderr_filename, parameter_dict)
                    stderr_filename = \
                        os.path.expandvars(os.path.expanduser(stderr_filename))
                else:
                    stderr_filename = "stderr"
                stderr_path = os.path.join(work_dir, stderr_filename)
                stderr = open(stderr_path, "a")

        # Use operation specific work directory
        if self._work_dir is not None and len(self._work_dir) > 0:
            new_work_dir = jube2.util.util.substitution(
                self._work_dir, parameter_dict)
            new_work_dir = os.path.expandvars(os.path.expanduser(new_work_dir))
            work_dir = os.path.join(work_dir, new_work_dir)
            # Create directory if it does not exist
            if not jube2.conf.DEBUG_MODE and not os.path.exists(work_dir):
                os.makedirs(work_dir)

        if not only_check_pending:

            abs_info_file_path = \
                os.path.abspath(os.path.join(work_dir,
                                             jube2.conf.ENVIRONMENT_INFO))

            # Select unix shell
            shell = jube2.conf.STANDARD_SHELL
            if "JUBE_EXEC_SHELL" in os.environ:
                alt_shell = os.environ["JUBE_EXEC_SHELL"].strip()
                if len(alt_shell) > 0:
                    shell = alt_shell

            # Execute "do"
            LOGGER.debug(">>> {0}".format(do))
            if (not jube2.conf.DEBUG_MODE) and (do != ""):
                LOGGER.debug("    stdout: {0}".format(
                    os.path.abspath(stdout_path)))
                LOGGER.debug("    stderr: {0}".format(
                    os.path.abspath(stderr_path)))
                try:
                    if jube2.conf.VERBOSE_LEVEL > 1:
                        stdout_handle = subprocess.PIPE
                    else:
                        stdout_handle = stdout
                    sub = subprocess.Popen(
                        [shell, "-c",
                         "{0} && env > \"{1}\"".format(do,
                                                       abs_info_file_path)],
                        cwd=work_dir, stdout=stdout_handle,
                        stderr=stderr, shell=False,
                        env=env)
                except OSError:
                    stdout.close()
                    stderr.close()
                    raise RuntimeError(("Error (returncode <> 0) while " +
                                        "running \"{0}\" in " +
                                        "directory \"{1}\"")
                                       .format(do, os.path.abspath(work_dir)))

                # stdout verbose output
                if jube2.conf.VERBOSE_LEVEL > 1:
                    while True:
                        read_out = sub.stdout.read(
                            jube2.conf.VERBOSE_STDOUT_READ_CHUNK_SIZE)
                        if (not read_out):
                            break
                        else:
                            print(read_out.decode(errors="ignore"), end="")
                            try:
                                stdout.write(read_out)
                            except TypeError:
                                stdout.write(read_out.decode(errors="ignore"))
                            time.sleep(jube2.conf.VERBOSE_STDOUT_POLL_SLEEP)
                    sub.communicate()

                returncode = sub.wait()

                # Close filehandles
                stdout.close()
                stderr.close()

                env = Operation.read_process_environment(work_dir)

                # Read and store new environment
                if (environment is not None) and (returncode == 0):
                    environment.clear()
                    environment.update(env)

                if returncode != 0:
                    if os.path.isfile(stderr_path):
                        stderr = open(stderr_path, "r")
                        stderr_msg = stderr.readlines()
                        stderr.close()
                    else:
                        stderr_msg = ""
                    try:
                        raise RuntimeError(
                            ("Error (returncode <> 0) while running \"{0}\" " +
                             "in directory \"{1}\"\nMessage in \"{2}\":" +
                             "{3}\n{4}").format(
                                do,
                                os.path.abspath(work_dir),
                                os.path.abspath(stderr_path),
                                "\n..." if len(stderr_msg) >
                                jube2.conf.ERROR_MSG_LINES else "",
                                "\n".join(stderr_msg[
                                    -jube2.conf.ERROR_MSG_LINES:])))
                    except UnicodeDecodeError:
                        raise RuntimeError(
                            ("Error (returncode <> 0) while running \"{0}\" " +
                             "in directory \"{1}\"").format(
                                do,
                                os.path.abspath(work_dir)))

        continue_op = True
        continue_cycle = True

        # Check if further execution was skipped
        if self._break_filename is not None:
            break_filename = jube2.util.util.substitution(
                self._break_filename, parameter_dict)
            break_filename = \
                os.path.expandvars(os.path.expanduser(break_filename))
            if os.path.exists(os.path.join(work_dir, break_filename)):
                LOGGER.debug(("\"{0}\" was found, workpackage execution and "
                              " further loop continuation was stopped.")
                             .format(break_filename))
                continue_cycle = False

        # Waiting to continue
        if self._async_filename is not None:
            async_filename = jube2.util.util.substitution(
                self._async_filename, parameter_dict)
            async_filename = \
                os.path.expandvars(os.path.expanduser(async_filename))
            if not os.path.exists(os.path.join(work_dir, async_filename)):
                LOGGER.debug("Waiting for file \"{0}\" ..."
                             .format(async_filename))
                if jube2.conf.DEBUG_MODE:
                    LOGGER.debug("  skip waiting")
                else:
                    continue_op = False

        return continue_op, continue_cycle

    def etree_repr(self):
        """Return etree object representation"""
        do_etree = ET.Element("do")
        do_etree.text = self._do
        if self._async_filename is not None:
            do_etree.attrib["done_file"] = self._async_filename
        if self._break_filename is not None:
            do_etree.attrib["break_file"] = self._break_filename
        if self._stdout_filename is not None:
            do_etree.attrib["stdout"] = self._stdout_filename
        if self._stderr_filename is not None:
            do_etree.attrib["stderr"] = self._stderr_filename
        if self._active != "true":
            do_etree.attrib["active"] = self._active
        if self._shared:
            do_etree.attrib["shared"] = "true"
        if self._work_dir is not None:
            do_etree.attrib["work_dir"] = self._work_dir
        return do_etree

    def __repr__(self):
        return self._do

    @staticmethod
    def read_process_environment(work_dir, remove_after_read=True):
        """Read standard environment info file in given directory."""
        env = dict()
        last = None
        env_file_path = os.path.join(work_dir, jube2.conf.ENVIRONMENT_INFO)
        if os.path.isfile(env_file_path):
            env_file = open(env_file_path, "r")
            for line in env_file:
                line = line.rstrip()
                matcher = re.match(r"^(\S.*?)=(.*?)$", line)
                if matcher:
                    env[matcher.group(1)] = matcher.group(2)
                    last = matcher.group(1)
                elif last is not None:
                    env[last] += "\n" + line
            env_file.close()
            if remove_after_read:
                os.remove(env_file_path)
        return env



# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Shell Completions"""

from __future__ import (print_function,
                        unicode_literals,
                        division)


import jube2.main


# This is formatted once.
BASH_CASE_TEMPLATE = """\
        "{command}")
            COMPREPLY=( $(compgen -W "{opts}" -- ${{cur}}) )
            return 0
            ;;
"""

# This is formatted once.
BASH_SCRIPT_TEMPLATE = """
_{command_name} ()
{{
    local cur prev words cword comm subparsers subcom iter

    COMPREPLY=()

    words=(${{COMP_WORDS[@]}})
    cword=COMP_CWORD
    comm=${{words[0]}}
    cur="${{words[cword]}}"
    prev="${{words[cword-1]}}"
    subcom="${{words[0]}}"
    for iter in ${{words[@]:1}}; do
        if [[ $iter != -* ]] && [[ " {all_subcoms} " == *" $iter "* ]]; then
            subcom=$iter
            break
        fi
    done
    subparsers="{subparser}"

    if [[ ${{cur}} == -* ]] ; then
        case "${{subcom}}" in
{cases_sub}
            *)
        esac
    elif [[ ${{subcom}} == "$comm" ]] ; then
        COMPREPLY=( $(compgen -W "${{subparsers}}" -- ${{cur}}) )
    fi
}} &&
complete -o bashdefault -o default -F _{command_name} {command_name}
"""


def complete_function_bash(args):
    """Print completion function for bash."""

    subparser = jube2.main.gen_subparser_conf()
    all_sub_names = " ".join(sorted(subparser))

    parser = sorted([opt
                     for opts, kwargs in jube2.main.gen_parser_conf()
                     for opt in opts if opt.startswith("--")])

    command_name = args.command_name[0]

    complete_options = dict()
    # Iterate over all subparsers
    for sub_name, sub in sorted(subparser.items()):
        if "arguments" not in sub:
            continue
        # Iterate over all their options
        tmp_list = [argument
                    for key in sub["arguments"]
                    for argument in key
                    if argument.startswith("--")]
        complete_options[sub_name] = " ".join(tmp_list)

    cases_sub = "".join(BASH_CASE_TEMPLATE.format(command=command,
                                                  opts=opts)
                        for command, opts in sorted(complete_options.items()))

    cases_sub += BASH_CASE_TEMPLATE.format(command=command_name,
                                           opts=" ".join(parser))

    subparser_str = " ".join(sorted(subparser.keys()))
    script = BASH_SCRIPT_TEMPLATE.format(
        subparser=subparser_str, cases_sub=cases_sub,
        command_name=command_name, all_subcoms=all_sub_names)
    print(script)



# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""ASCII and string output generators"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

import jube2.conf
import textwrap
import copy
import sys
import xml.etree.ElementTree as ET


def text_boxed(text):
    """Create an ASCII boxed version of text."""
    box = "#" * jube2.conf.DEFAULT_WIDTH
    for line in text.split("\n"):
        box += "\n"
        lines = ["# {0}".format(element) for element in
                 textwrap.wrap(line.strip(), jube2.conf.DEFAULT_WIDTH - 2)]
        if len(lines) == 0:
            box += "#"
        else:
            box += "\n".join(lines)
    box += "\n" + "#" * jube2.conf.DEFAULT_WIDTH
    return box


def text_line():
    """Return a horizonal ASCII line"""
    return "#" * jube2.conf.DEFAULT_WIDTH


def text_table(entries_ext, use_header_line=False, indent=1, align_right=True,
               auto_linebreak=True, colw=None, pretty=True, separator=None,
               transpose=False):
    """Create a ASCII based table.
    entries must contain a list of lists, use_header_line can be used to
    mark the first entry as title.

    Return the ASCII table
    """

    if not pretty:
        auto_linebreak = False
        use_header_line = False
        indent = 0

    # Transpose data entries if needed
    if transpose:
        entries = list(zip(*entries_ext))
        use_header_line = False
    else:
        entries = copy.deepcopy(entries_ext)

    max_length = list()
    table_str = ""
    header_line_used = not use_header_line

    # calculate needed maxlength
    for item in entries:
        for i, text in enumerate(item):
            if i > len(max_length) - 1:
                max_length.append(0)
            if pretty:
                for line in text.splitlines():
                    max_length[i] = max(max_length[i], len(line))
                if auto_linebreak:
                    max_length[i] = min(max_length[i],
                                        jube2.conf.MAX_TABLE_CELL_WIDTH)

    if colw is not None:
        for i, maxl in enumerate(max_length):
            if i < len(colw):
                max_length[i] = max(maxl, colw[i])

    # fill cells
    for item in entries:

        # Wrap text
        wraps = list()
        for text in item:
            if auto_linebreak:
                lines = list()
                for line in text.splitlines():
                    lines += \
                        textwrap.wrap(line, jube2.conf.MAX_TABLE_CELL_WIDTH)
                wraps.append(lines)
            else:
                if pretty:
                    wraps.append(text.splitlines())
                else:
                    wraps.append([text.replace("\n", " ")])

        grow = True
        height = 0
        while grow:
            grow = False
            line_str = " " * indent
            for i, wrap in enumerate(wraps):
                grow = grow or len(wrap) > height + 1
                if len(wrap) > height:
                    text = wrap[height]
                else:
                    text = ""
                if align_right and height == 0:
                    align = ">"
                else:
                    align = "<"
                line_str += \
                    ("{0:" + align + str(max_length[i]) + "s}").format(text)
                if pretty:
                    if i < len(max_length) - 1:
                        if separator is None:
                            line_str += " | "
                        else:
                            line_str += separator
                else:
                    if i < len(max_length) - 1:
                        if separator is None:
                            line_str += ","
                        else:
                            line_str += separator
            line_str += "\n"
            table_str += line_str
            height += 1

        if not header_line_used:
            # Create title separator line
            table_str += " " * indent
            for i, cell_length in enumerate(max_length):
                table_str += "-" * cell_length
                if i < len(max_length) - 1:
                    table_str += "-+-"
            table_str += "\n"
            header_line_used = True
    return table_str


def print_loading_bar(current_cnt, all_cnt, wait_cnt=0, error_cnt=0):
    """Show a simple loading animation"""
    width = jube2.conf.DEFAULT_WIDTH - 10
    cnt = dict()
    if all_cnt > 0:
        cnt["done_cnt"] = (current_cnt * width) // all_cnt
        cnt["wait_cnt"] = (wait_cnt * width) // all_cnt
        cnt["error_cnt"] = (error_cnt * width) // all_cnt
    else:
        cnt["done_cnt"] = 0
        cnt["wait_cnt"] = 0
        cnt["error_cnt"] = 0

    # shrink cnt if there was some rounding issue
    for key in ("wait_cnt", "error_cnt"):
        if (cnt[key] > 0) and (width < sum(cnt.values())):
            cnt[key] = max(0, width - sum([cnt[k] for k in cnt if k != key]))

    # fill up medium_cnt if there was some rounding issue
    if (current_cnt + wait_cnt + error_cnt == all_cnt) and \
            (sum(cnt.values()) < width):
        for key in ("wait_cnt", "error_cnt", "done_cnt"):
            if cnt[key] > 0:
                cnt[key] += width - sum(cnt.values())
                break

    cnt["todo_cnt"] = width - sum(cnt.values())

    bar_str = "\r{0}{1}{2}{3} ({4:3d}/{5:3d})".format("#" * cnt["done_cnt"],
                                                      "0" * cnt["wait_cnt"],
                                                      "E" * cnt["error_cnt"],
                                                      "." * cnt["todo_cnt"],
                                                      current_cnt, all_cnt)
    sys.stdout.write(bar_str)
    sys.stdout.flush()


def element_tree_tostring(element, encoding=None):
    """A more encoding friendly ElementTree.tostring method"""
    class Dummy(object):

        """Dummy class to offer write method for etree."""

        def __init__(self):
            self._data = list()

        @property
        def data(self):
            """Return data"""
            return self._data

        def write(self, *args):
            """Simulate write"""
            self._data.append(*args)
    file_dummy = Dummy()
    ET.ElementTree(element).write(file_dummy, encoding)
    return "".join(dat.decode(encoding) for dat in file_dummy.data)


def format_value(format_string, value):
    """Return formated value"""
    if (type(value) is not int) and \
            (("d" in format_string) or ("b" in format_string) or
             ("c" in format_string) or ("o" in format_string) or
             ("x" in format_string) or ("X" in format_string)):
        value = int(float(value))
    elif (type(value) is not float) and \
         (("e" in format_string) or ("E" in format_string) or
          ("f" in format_string) or ("F" in format_string) or
          ("g" in format_string) or ("G" in format_string)):
        value = float(value)
    format_string = "{{0:{0}}}".format(format_string)
    return format_string.format(value)



# JUBE Benchmarking Environment
# Copyright (C) 2008-2018
# Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre
# http://www.fz-juelich.de/jsc/jube
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Syslogtype definition"""

from __future__ import (print_function,
                        unicode_literals,
                        division)

from jube2.result_types.keyvaluesresult import KeyValuesResult
from jube2.result import Result
import xml.etree.ElementTree as ET
import jube2.log
import jube2.conf
import logging.handlers

LOGGER = jube2.log.get_logger(__name__)


class SysloggedResult(KeyValuesResult):

    """A result that gets sent to syslog."""

    class SyslogData(KeyValuesResult.KeyValuesData):

        """Table data"""

        def __init__(self, name_or_other, syslog_address=None,
                     syslog_host=None, syslog_port=None,
                     syslog_fmt_string=None):
            if type(name_or_other) is KeyValuesResult.KeyValuesData:
                self._name = name_or_other.name
                self._keys = name_or_other.keys
                self._data = name_or_other.data
                self._benchmark_ids = name_or_other.benchmark_ids
            else:
                KeyValuesResult.KeyValuesData.__init__(self, name_or_other)
            self._syslog_address = syslog_address
            self._syslog_host = syslog_host
            self._syslog_port = syslog_port
            self._syslog_fmt_string = syslog_fmt_string

        def create_result(self, show=True, filename=None, **kwargs):
            """Create result output"""
            # If there are multiple benchmarks, add benchmark id information
            if len(set(self._benchmark_ids)) > 1:
                self.add_id_information(reverse=kwargs.get("reverse", False))

            if self._syslog_address is not None:
                address = self._syslog_address
            else:
                address = (self._syslog_host, self._syslog_port)

            handler = logging.handlers.SysLogHandler(
                address=address,
                facility=logging.handlers.SysLogHandler.LOG_USER
            )
            handler.setFormatter(logging.Formatter(
                fmt=self._syslog_fmt_string))

            # get logger
            log = logging.getLogger("jube")
            log.setLevel(logging.INFO)
            log.addHandler(handler)

            # create log output
            for dataset in self.data:
                entry = list()
                for i, key in enumerate(self.keys):
                    entry.append("{0}={1}".format(key.name, dataset[i]))
                # Log result
                if show:
                    if not jube2.conf.DEBUG_MODE:
                        log.info(" ".join(entry))
                    LOGGER.debug("Logged: {0}\n".format(" ".join(entry)))

            # remove handler to avoid double logging
            log.removeHandler(handler)

    def __init__(self, name, syslog_address=None, syslog_host=None,
                 syslog_port=None, syslog_fmt_string=None, sort_names=None,
                 res_filter=None):
        KeyValuesResult.__init__(self, name, sort_names, res_filter)
        if (syslog_address is None) and (syslog_host is None) and \
                (syslog_port is None):
            raise IOError("Neither a syslog address nor a hostname port " +
                          "combination specified.")
        if (syslog_host is not None) and (syslog_address is not None):
            raise IOError("Please specify a syslog address or a hostname, " +
                          "not both at the same time.")
        if (syslog_host is not None) and (syslog_port is None):
            self._syslog_port = 514
        self._syslog_address = syslog_address
        self._syslog_host = syslog_host
        self._syslog_port = syslog_port
        if syslog_fmt_string is None:
            self._syslog_fmt_string = jube2.conf.SYSLOG_FMT_STRING
        else:
            self._syslog_fmt_string = syslog_fmt_string

    def create_result_data(self):
        """Create result data"""
        result_data = KeyValuesResult.create_result_data(self)
        return SysloggedResult.SyslogData(result_data, self._syslog_address,
                                          self._syslog_host, self._syslog_port,
                                          self._syslog_fmt_string)

    def etree_repr(self):
        """Return etree object representation"""
        result_etree = Result.etree_repr(self)
        syslog_etree = ET.SubElement(result_etree, "syslog")
        syslog_etree.attrib["name"] = self._name
        if self._syslog_address is not None:
            syslog_etree.attrib["address"] = self._syslog_address
        if self._syslog_host is not None:
            syslog_etree.attrib["host"] = self._syslog_host
        if self._syslog_port is not None:
            syslog_etree.attrib["port"] = self._syslog_port
        if self._syslog_fmt_string is not None:
            syslog_etree.attrib["format"] = self._syslog_fmt_string
        if self._res_filter is not None:
            syslog_etree.attrib["filter"] = self._res_filter
        if len(self._sort_names) > 0:
            syslog_etree.attrib["sort"] = \
                jube2.conf.DEFAULT_SEPARATOR.join(self._sort_names)
        for key in self._keys:
            syslog_etree.append(key.etree_repr())
        return result_etree
